"""This module provides the `TWDA` singleton: Tournament Wining Decks Archive.

If it has not been initialized, TWDA will evaluate to False.
TWDA must be configured with `TWDA.configure()` before being used.
"""
import collections
import itertools
import logging
import pickle
import re

import arrow
from arrow.parser import ParserError

from . import config
from . import deck
from . import vtes

logger = logging.getLogger()


class AnalysisError(Exception):
    pass


# Headers used in TWDA, e.g. "Action modifiers/Combat cards total: (4)"
HEADERS = set()
HEADERS.update(config.HEADERS)
HEADERS.update([h + "s" for h in config.HEADERS] + ["ally", "allies"])
HEADERS.update(
    ["{}/{}".format(h1, h2) for h1, h2 in itertools.permutations(HEADERS, 2)]
)
HEADERS.update(config.ADDITIONAL_HEADERS)
HEADERS.update([h + " card" for h in HEADERS] + [h + " cards" for h in HEADERS])
HEADERS.update(["card", "cards", "deck"])
HEADERS.update([h + " total" for h in HEADERS])
HEADERS.update([h + ":" for h in HEADERS])


class _TWDA(collections.OrderedDict):
    """ An OrderedDict of the TWDA. Parsing TWDA.html is the hard part.
    """

    def load_html(self, source, binary=False):
        """Load from TWDA.html
        """
        for lines, twda_id in self._get_decks_html(source, binary):
            try:
                self[twda_id] = self._load_deck_html(lines, twda_id)
            except ParserError as e:
                logger.error(e)
        logger.info("TWDA loaded")
        pickle.dump(TWDA, open(config.TWDA_FILE, "wb"))

    def configure(self, date_from=None, date_to=None, min_players=0):
        """Compute `self.spoilers`: cards played in over 25% of decks
        """
        if date_from:
            for key in [key for key, value in self.items() if value.date < date_from]:
                del self[key]
        if date_to:
            for key in [key for key, value in self.items() if value.date >= date_to]:
                del self[key]
        if min_players:
            for key in [
                key
                for key, value in self.items()
                if value.players_count and value.players_count < min_players
            ]:
                del self[key]
        if len(self) > 50:
            self.spoilers = {
                name
                for name, count in collections.Counter(
                    itertools.chain.from_iterable(d.keys() for d in self.values())
                ).items()
                if count > len(self) / 4
            }
            logger.debug("Spoilers: {}".format(self.spoilers))
        else:
            self.spoilers = {}

    def _get_decks_html(self, source, binary):
        """Get lines for each deck, using HTML tags as separators.
        """
        lines = []
        for line_num, line in enumerate(source.readlines(), start=1):
            if binary:
                line = line.decode("utf-8")
            try:
                twda_id = re.match(r"^<a id=([^\s]*)\s", line).group(1)
            except AttributeError:
                pass
            if re.match(r"^<hr><pre>$", line):
                lines = []
            elif re.match(r"^</pre>", line):
                yield lines, twda_id
            else:
                lines.append((line_num, line[:-1]))

    def _load_deck_html(self, lines, twda_id):
        """Parse TWDA.html lines to build a deck.
        """
        current = deck.Deck()
        headers = False
        # First three lines are stable throughout TWDA
        current.event = lines[0][1]
        current.place = lines[1][1]
        current.date = arrow.get(lines[2][1], "MMMM Do YYYY")
        start = 3
        # Score and players count are not always set
        try:
            current.score = re.match(r"^(\d+)R\+F", lines[start][1]).group(1)
            start += 1
        except AttributeError:
            pass
        try:
            current.players_count = int(
                re.match(r"^(\d+)\s*player", lines[start][1]).group(1)
            )
            start += 1
        except AttributeError:
            pass
        # Player is always set, but after score and count if any
        current.player = lines[start][1]
        start += 1
        # After this header, comments and format tricks happen
        comment = None
        for line_num, line in lines[start + 1 :]:
            # handle C-style multiline comments (very common) /* comment */
            if "*/" in line:
                comment = None
                continue
            if comment:
                continue
            if "/*" in line:
                line, comment = line.split("/*")
                if comment and "*/" in comment:
                    comment = None
                else:
                    comment = True
            if not line:
                continue
            # handle separtor lines
            if set(line).issubset({"=", "-"}):
                headers = True
                continue
            # monoline [comment] is also common
            if line[0] == "[" and line[-1] == "]":
                continue
            # "meaningful content -- comment" is widespread
            line = line.split("--", 1)[0]
            # as is C-style "meaningful content //comment"
            line = line.split("//", 1)[0]
            # remove misplaced spaces around punctuation for easy parsing
            line = line.replace(" :", ":")
            line = line.replace("( ", "(")
            line = line.replace(" )", ")")
            line = line.replace(" /", "/")
            line = line.replace("/ ", "/")
            # at this point we may not have a meaningful line anymore
            if not line:
                continue
            # if we did not hit any header (e.g. "Crypt"), try to find
            # deck name, author and comments
            if not headers:
                try:
                    current.name = re.match(
                        r"^\s*(?:d|D)eck\s?(?:n|N)ame\s*:\s*(.*)$", line
                    ).group(1)
                    continue
                except AttributeError:
                    pass
                try:
                    current.author = re.match(
                        r"(?:(?:(?:c|C)reated)|(?:(?:d|D)eck))"
                        r"\s*(?:b|B)y\s*:?\s*(.*)$",
                        line,
                    ).group(1)
                    continue
                except AttributeError:
                    pass
                try:
                    line = re.match(
                        r"^(?:(?:d|D)eck)?"
                        r"\s*(?:(?:(?:d|D)escription)|(?:(?:n|N)otes?))\s*"
                        r":?\s*(.*)$",
                        line,
                    ).group(1)
                except AttributeError:
                    pass
                if not line:
                    continue
                # if we hit the "Crypt" marker, deck list is beginning
                if not re.match(r"^(c|C)rypt", line):
                    # more often than not, first comment line is the deck name
                    # they tend to be relatively short, though
                    if not current.name and len(line) < 70:
                        current.name = line
                        continue
                    current.comments += line + "\n"
                    continue
            # lower all chars for easier parsing
            line = line.lower()
            name, count = _get_card(line)
            if name and count:
                if name.startswith("<a "):
                    continue
                if name in HEADERS:
                    continue
                if not headers and not current:
                    continue
                try:
                    card = vtes.VTES[name]
                except KeyError:
                    logger.error(
                        "[{:<6}] unknown card [{}] - [{}]".format(line_num, name, line)
                    )
                    continue
                current.update({card["Name"]: count})
            else:
                logger.error("[{:<6}] no card matched [{}]".format(line_num, line))
        library_count = current.cards_count(vtes.VTES.is_library)
        if library_count < 60:
            logger.info(
                "[{:<6}] Deck #{} is missing library cards [{}]".format(
                    line_num, twda_id, current
                )
            )
        if library_count > 90:
            logger.info(
                "[{:<6}] Deck #{} has too many cards [{}]".format(
                    line_num, twda_id, current
                )
            )
        if current.cards_count(vtes.VTES.is_crypt) < 12:
            logger.info(
                "[{:<6}] Deck #{} is missing crypt cards [{}]".format(
                    line_num, twda_id, current
                )
            )
        return current

    def no_spoil(self, card):
        return not getattr(self, "spoilers", None) or card not in self.spoilers


def _get_card(line):
    """Find card name and count in a TWDA.html line

    There are a lot of circumvoluted cases to take into account, cf. tests.
    """
    card_match = re.match(
        r"^\s*(?:-|_)?(?:(?:x|X|\*|_)?\s*"
        r"(?P<count>\d{1,2})(?!(?:(?:st)|(?:nd)|(?:rd)|(?:th))))?"
        r"\s*(?:(?:x|X|-)\s)?\*?\s*"
        r"(?P<name>(?:(?:channel 10)|.+?))"
        r"(?:(?:"
        r"(?:\s|\(|\[|:|/|=)+(?:-|x|\*|\s)*"
        r"(?P<count_or_cap>\d{1,3})"
        r"(?!(?:(?:st)|(?:nd)|(?:rd)|(?:th)|(?:.?\d)|b|p|(?:..?port)))"
        r")|$)",
        line,
    )
    if not card_match:
        return None, None
    name = card_match.group("name").strip()
    count = int(card_match.group("count") or card_match.group("count_or_cap") or 1)
    return name, count


try:
    TWDA = pickle.load(open(config.TWDA_FILE, "rb"))
except (FileNotFoundError, EOFError):
    TWDA = _TWDA()  # evaluates to False as it is empty



"""This module provides the `VTES` singleton: the VTES cards library.

If it has not been initialized, VTES will evaluate to False.
VTES must be configured with `VTES.configure()` before being used.
"""
import csv
import functools
import difflib
import itertools
import logging
import pickle

from . import config


logger = logging.getLogger()


class _VTES(dict):
    """VTES cards database that matches incomplete or misspelled card names.

    Keys are lower case card names.
    """

    def __getitem__(self, key):
        """Get a card, try to find a good matching.

        Use lowercase only, config.REMAP for common errors and abbreviations.
        Use difflib to match incomplete or misspelled names
        """
        key = key.lower()
        try:
            return super().__getitem__(key)
        except KeyError:
            # REMAP must match exactly
            if key in config.REMAP:
                return self[config.REMAP[key]]
            match = self._fuzzy_match(key)
            if match:
                return self[match]
            try:
                while key:
                    key, _ = key.rsplit(maxsplit=1)
                    if key in self:
                        return self[key]
            except ValueError:
                pass
            raise

    def __setitem__(self, key, value):
        return super().__setitem__(key.lower(), value)

    def __delitem__(self, key, value):
        return super().__setitem__(key.lower(), value)

    def __contains__(self, item):
        return super().__contains__(item.lower())

    def configure(self):
        """Add alternative names for some cards
        """
        # note we difflib match on AKA but not on REMAP.
        for alternative, card in config.AKA.items():
            self[alternative] = self[card]

    def __hash__(self):
        return id(self)

    @functools.lru_cache()
    def _fuzzy_match(self, name):
        """Use a cache to accelerate matching of common misspellings.
        """
        # 0.8 is an empirical value, seems to work nice
        result = difflib.get_close_matches(name, self.keys(), n=1, cutoff=0.8)
        if result:
            match = result[0]
            logger.info("misspelled [{}] matched [{}]".format(name, match))
            return match

    def trait_choices(self, trait):
        """Get all registered values for a given card trait.

        Empty values are removed.

        >>> sorted(VTES.trait_choices('Banned'))
        ['1995', '1997', '1999', '2005', '2008', '2013']
        """

        def get_traits(trait_value):
            if "/" in trait_value:
                for value in trait_value.split("/"):
                    yield value.strip()
            elif "&" in trait_value:
                yield "Combo"
            else:
                yield trait_value

        return set(
            itertools.chain.from_iterable(
                get_traits(card[trait]) for card in self.values() if card.get(trait)
            )
        )

    @property
    def disciplines(self):
        return sorted(self.trait_choices("Discipline"))

    @property
    def types(self):
        return sorted(self.trait_choices("Type"))

    @property
    def clans(self):
        return sorted(self.trait_choices("Clan"))

    def load_csv(self, stream):
        for card in csv.DictReader(stream):
            if card.get("Adv"):
                name = card["Name"] + " (ADV)"
            else:
                name = card["Name"]
            self[name] = card
            if card.get("Aka"):
                self[card["Aka"]] = card
                if card["Aka"][-5:] == ", The":
                    self[("The " + card["Name"][:-5])] = card
            if ", The" in name:
                name = name.split(", The")[0]
                self["The " + name] = card
                self[name] = card
            # add name without suffix if any
            # ':' can not be seen as a suffix marker, e.g. Praxis Seizure:
            if "," in name:
                alternative = name.split(",")[0]
                if len(alternative) > 6:  # e.g. Pier 13, Port of Baltimore
                    self[alternative] = card
            if "(" in card["Name"]:
                alternative = name.split("(")[0].strip()
                # e.g. Pentex(TM) Subversion
                if len(alternative) > 6:
                    self[alternative] = card

        # pickle this
        pickle.dump(VTES, open(config.VTES_FILE, "wb"))

    # usual traits selectors
    def is_crypt(self, card):
        return VTES[card]["Type"] in ("Vampire", "Imbued")

    def is_library(self, card):
        return VTES[card]["Type"] not in ("Vampire", "Imbued")

    # handle multi valued traits
    def is_disc(self, card, discipline):
        if discipline == "Combo":
            return "&" in VTES[card].get("Discipline", "")
        return {discipline} & {
            disc.strip()
            for d in VTES[card].get("Discipline", "").split("/")
            for disc in d.split("&")
        }

    def is_type(self, card, type_):
        return {type_} & {t.strip() for t in VTES[card]["Type"].split("/")}

    def is_clan(self, card, clan):
        return {clan} & {t.strip() for t in VTES[card]["Clan"].split("/")}

    # a consistent deck display matching our parsing rules of TWDA.html
    def deck_to_txt(self, deck):
        def _type(card):
            return config.TYPE_ORDER.index(self[card[0]]["Type"])

        lines = []
        if deck.event:
            lines.append(deck.event)
        if deck.place:
            lines.append(deck.place)
        if deck.date:
            lines.append(deck.date.format("MMMM Do YYYY"))
        if deck.score:
            lines.append(deck.score + "R+F")
        if deck.players_count:
            lines.append("{} players".format(deck.players_count))
        if deck.player:
            lines.append(deck.player)
        lines.append("")
        if deck.name:
            lines.append("Deck Name: " + deck.name)
        if deck.author:
            lines.append("Created by: " + deck.author)
        if deck.comments:
            lines.append("Description:")
            lines.append(deck.comments)
        lines.append("")
        lines.append("-- Crypt: ({} cards)".format(deck.cards_count(self.is_crypt)))
        lines.append("---------------------------------------")
        for card, count in deck.cards(self.is_crypt):
            lines.append(
                "{:<2} {:<35} {:<2} {:<25} {}:{}".format(
                    count,
                    card + (" (ADV)" if self[card]["Adv"] else ""),
                    self[card]["Capacity"],
                    self[card]["Disciplines"],
                    self[card]["Clan"],
                    self[card]["Group"],
                )
            )
        lines.append("-- Library ({})".format(deck.cards_count(self.is_library)))
        # sort by type, count (dec.), name
        library_cards = sorted(
            deck.cards(self.is_library), key=lambda a: (_type(a), -a[1], a[0])
        )
        for kind, cards in itertools.groupby(library_cards, key=_type):
            c1, c2 = itertools.tee(cards)
            lines.append(
                "-- {} ({})".format(
                    config.TYPE_ORDER[kind], sum(count for card, count in c1)
                )
            )
            for card, count in c2:
                lines.append("{:<2} {}".format(count, card))
        return "\n".join(lines)


try:
    VTES = pickle.load(open(config.VTES_FILE, "rb"))
except (FileNotFoundError, EOFError):
    VTES = _VTES()  # evaluates to False as it is empty



"""Deck class: pickling and card access under conditions.
"""
import collections

import arrow
import arrow.parser


class Deck(collections.Counter):
    """A VTES deck, including meta information such as author and decription.

    Meta information:
        - name: Name of the deck
        - author: Deck author
        - comments: Comments on the deck
        - event: Event (for TWD)
        - place: Place where the event was held
        - date: Date on which the event was held
        - score: Score of the deck at the event
        - players_count: Count of players at the event
        - player: Player who played the deck
    """

    def __init__(self, *args, **kwargs):
        self.author = kwargs.pop("author", None)
        super().__init__(*args, **kwargs)
        self.event = None
        self.place = None
        self.date = None
        self.score = None
        self.players_count = 0
        self.player = None
        self.name = None
        self.comments = ""

    def __getstate__(self):
        """For pickle serialization: Deck inherits dict and their special
        handling of pickle.
        """
        return {
            "cards": collections.OrderedDict(self.cards()),
            "author": self.author,
            "event": self.event,
            "place": self.place,
            "date": self.date.format("MMMM Do YYYY") if self.date else None,
            "score": self.score,
            "players_count": self.players_count,
            "player": self.player,
            "name": self.name,
            "comments": self.comments,
        }

    def __setstate__(self, state):
        """For pickle deserialization.
        """
        self.author = state.get("author")
        self.event = state.get("event")
        self.place = state.get("place")
        try:
            self.date = arrow.get(state.get("date"), "MMMM Do YYYY")
        except arrow.parser.ParserError:
            pass
        self.score = state.get("score")
        self.players_count = int(state.get("players_count", 0))
        self.player = state.get("player")
        self.name = state.get("name")
        self.comments = state.get("comments", "")
        self.update(state.get("cards", {}))

    def __reduce__(self):
        """For pickle.
        """
        return (Deck, (), self.__getstate__())

    def __str__(self):
        return self.name or "(No Name)"

    def cards(self, condition=None):
        """Generator yielding (card_name, count), with an optional filter.
        """
        for card, count in self.items():
            if condition and not condition(card):
                continue
            yield card, count

    def card_names(self, condition=None):
        """Generator yielding card names with an optional filter.
        """
        for card, count in self.cards(condition):
            yield card

    def cards_count(self, condition=None):
        """Card counts with an optional filter.
        """
        return sum(count for card, count in self.cards(condition))



"""Configuration.
"""
from os import path

VEKN_TWDA_URL = "http://www.vekn.fr/decks/twd.htm"
VEKN_VTES_URL = "http://www.vekn.net/images/stories/downloads/vtescsv_utf8.zip"
VEKN_VTES_LIBRARY_FILENAME = "vteslib.csv"
VEKN_VTES_CRYPT_FILENAME = "vtescrypt.csv"
TWDA_FILE = path.join(path.dirname(__file__), "TWDA.pkl")
VTES_FILE = path.join(path.dirname(__file__), "VTES.pkl")
HEADERS = [
    "master",
    "minion",
    "crypt",
    "library",
    "action",
    "combat",
    "conviction",
    "reaction",
    "retainer",
    "equipment",
    "equip",
    "event",
    "political action",
    "political",
    "politic",
    "power",
    "vote",
    "action modifier",
    "action mod",
    "modifier",
    "mod",
]
ADDITIONAL_HEADERS = [
    "combo",
    "combos",
    "multitype",
    "multitypes",
    "mixed",
    "miscellaneous",
    "misc",
    "average",
    "other",
    "others",
    "double",
    "doubles",
    "burn option",
    "equip",
    "total",
]
AKA = {
    # due to our card parsing
    "bang nakh": "Bang Nakh -- Tiger's Claws",
    # edition changes
    "raven": "Camille Devereux, The Raven",
    "gangrel de-evolution": "Gangrel Atavism",
    "mask of 1,000 faces": "Mask of a Thousand Faces",
    "carlton van wyk (hunter)": "Carlton Van Wyk",
}
REMAP = {
    # traditions - players really do whatever they want
    "the first tradition": "First Tradition: The Masquerade",
    "first tradition": "First Tradition: The Masquerade",
    "1st tradition": "First Tradition: The Masquerade",
    "the second tradition": "Second Tradition: Domain",
    "second tradition": "Second Tradition: Domain",
    "2nd trad": "Second Tradition: Domain",
    "2nd tradition": "Second Tradition: Domain",
    "2nd tradition: domain": "Second Tradition: Domain",
    "the third tradition": "Third Tradition: Progeny",
    "third tradition": "Third Tradition: Progeny",
    "3rd tradition": "Third Tradition: Progeny",
    "the fourth tradition": "Fourth Tradition: The Accounting",
    "fourth tradition": "Fourth Tradition: The Accounting",
    "4th tradition": "Fourth Tradition: The Accounting",
    "the fifth tradition": "Fifth Tradition: Hospitality",
    "fifth tradition": "Fifth Tradition: Hospitality",
    "5th tradition": "Fifth Tradition: Hospitality",
    "the sixth tradition": "Sixth Tradition: Destruction",
    "sixth tradition": "Sixth Tradition: Destruction",
    "6th tradition": "Sixth Tradition: Destruction",
    # hunting grounds
    "academic hg": "Academic Hunting Ground",
    "amusement park hg": "Amusement Park Hunting Ground",
    "asylum hg": "Asylum Hunting Ground",
    "base hg": "Base Hunting Ground",
    "campground hg": "Campground Hunting Ground",
    "corporate hg": "Corporate Hunting Ground",
    "fetish club hg": "Fetish Club Hunting Ground",
    "fetish hg": "Fetish Club Hunting Ground",
    "institution hg": "Institution Hunting Ground",
    "jungle hg": "Jungle Hunting Ground",
    "library hg": "Library Hunting Ground",
    "morgue hg": "Morgue Hunting Ground",
    "palace hg": "Palace Hunting Ground",
    "park hg": "Park Hunting Ground",
    "poacher's hg": "Poacher's Hunting Ground",
    "political hg": "Political Hunting Ground",
    "port hg": "Port Hunting Ground",
    "shanty Town hg": "Shanty Town Hunting Ground",
    "slum hg": "Slum Hunting Ground",
    "society hg": "Society Hunting Ground",
    "temple hg": "Temple Hunting Ground",
    "underworld hg": "Underworld Hunting Ground",
    "university hg": "University Hunting Ground",
    "uptown hg": "Uptown Hunting Ground",
    "warzone hg": "Warzone Hunting Ground",
    "zoo hg": "Zoo Hunting Ground",
    "academic h.g.": "Academic Hunting Ground",
    "amusement park h.g.": "Amusement Park Hunting Ground",
    "asylum h.g.": "Asylum Hunting Ground",
    "base h.g.": "Base Hunting Ground",
    "campground h.g.": "Campground Hunting Ground",
    "corporate h.g.": "Corporate Hunting Ground",
    "fetish club h.g.": "Fetish Club Hunting Ground",
    "institution h.g.": "Institution Hunting Ground",
    "jungle h.g.": "Jungle Hunting Ground",
    "library h.g.": "Library Hunting Ground",
    "morgue h.g.": "Morgue Hunting Ground",
    "palace h.g.": "Palace Hunting Ground",
    "park h.g.": "Park Hunting Ground",
    "poacher's h.g.": "Poacher's Hunting Ground",
    "political h.g.": "Political Hunting Ground",
    "port h.g.": "Port Hunting Ground",
    "shanty Town h.g.": "Shanty Town Hunting Ground",
    "slum h.g.": "Slum Hunting Ground",
    "society h.g.": "Society Hunting Ground",
    "temple h.g.": "Temple Hunting Ground",
    "underworld h.g.": "Underworld Hunting Ground",
    "university h.g.": "University Hunting Ground",
    "uptown h.g.": "Uptown Hunting Ground",
    "warzone h.g.": "Warzone Hunting Ground",
    "zoo h.g.": "Zoo Hunting Ground",
    "acad. hg": "Academic Hunting Ground",
    "univ. hg": "University Hunting Ground",
    "univ hg": "University Hunting Ground",
    # powerbases
    "pb: barranquilla": "Powerbase: Barranquilla",
    "pb: berlin": "Powerbase: Berlin",
    "pb: cape verde": "Powerbase: Cape Verde",
    "pb: chicago": "Powerbase: Chicago",
    "pb: los angeles": "Powerbase: Los Angeles",
    "pb: luanda": "Powerbase: Luanda",
    "pb: madrid": "Powerbase: Madrid",
    "pb: mexico city": "Powerbase: Mexico City",
    "pb: montreal": "Powerbase: Montreal",
    "pb:montreal": "Powerbase: Montreal",
    "pb: new york": "Powerbase: New York",
    "pb: rome": "Powerbase: Rome",
    "pb: savannah": "Powerbase: Savannah",
    "pb: tshwane": "Powerbase: Tshwane",
    "pb: washington, d.c.": "Powerbase: Washington, D.C.",
    "pb: zurich": "Powerbase: Zürich",
    "pb barranquilla": "Powerbase: Barranquilla",
    "pb berlin": "Powerbase: Berlin",
    "pb cape verde": "Powerbase: Cape Verde",
    "pb chicago": "Powerbase: Chicago",
    "pb los angeles": "Powerbase: Los Angeles",
    "pb luanda": "Powerbase: Luanda",
    "pb madrid": "Powerbase: Madrid",
    "pb mexico city": "Powerbase: Mexico City",
    "pb montreal": "Powerbase: Montreal",
    "pb new york": "Powerbase: New York",
    "pb rome": "Powerbase: Rome",
    "pb savannah": "Powerbase: Savannah",
    "pb tshwane": "Powerbase: Tshwane",
    "pb washington, d.c.": "Powerbase: Washington, D.C.",
    "pb zurich": "Powerbase: Zürich",
    # punctuation
    "behind you": "Behind You!",
    "psyche": "Psyche!",
    # known abbreviations
    "anthelios": "Anthelios, The Red Star",
    "anthelios, the": "Anthelios, The Red Star",
    "carlton": "Carlton Van Wyk",
    "con ag": "Conservative Agitation",
    "direct": "Direct Intervention",
    "delaying": "Delaying Tactics",
    "dreams": "Dreams of the Sphinx",
    "effective": "Effective Management",
    "elysium": "Elysium: The Arboretum",
    "elysium: versailles": "Elysium: The Palace of Versailles",
    "entice": "Enticement",
    "felix fix": 'Felix "Fix" Hessian (Wraith)',
    "forced": "Forced Awakening",
    "foreshadowing": "Foreshadowing Destruction",
    "golconda": "Golconda: Inner Peace",
    "govern": "Govern the Unaligned",
    "gtu": "Govern the Unaligned",
    "heidelberg": "Heidelberg Castle, Germany",
    "heidelburg": "Heidelberg Castle, Germany",
    "info highway": "Information Highway",
    "info hwy": "Information Highway",
    "js simmons": "J. S. Simmons, Esq.",
    "krc": "Kine Resources Contested",
    "krcg": "KRCG News Radio",
    "krcg news": "KRCG News Radio",
    "laptop": "Laptop Computer",
    "laptops": "Laptop Computer",
    "legal manip": "Legal Manipulations",
    "london eve star": "London Evening Star, Tabloid Newspaper",
    "mister winthrop": "Mr. Winthrop",
    "molotov": "Molotov Cocktail",
    "owl": "Owl Companion",
    "patagia: flaps": "Patagia: Flaps Allowing Limited Flight",
    "pentex": "Pentex(TM) Subversion",
    "storage": "Storage Annex",
    "sudden": "Sudden Reversal",
    "the rumor mill": "Rumor Mill, Tabloid Newspaper, The",
    "ventrue hq": "Ventrue Headquarters",
    "voter cap": "Voter Captivation",
    "wake with ef": "Wake with Evening's Freshness",
    "wake wef": "Wake with Evening's Freshness",
    "wwef": "Wake with Evening's Freshness",
    "wake...": "Wake with Evening's Freshness",
    "wake w/evening...": "Wake with Evening's Freshness",
    "wake": "Wake with Evening's Freshness",
    "wakes": "Wake with Evening's Freshness",
    "wakeys": "Wake with Evening's Freshness",
    "wmrh": "WMRH Talk Radio",
    # misspellings not fixed by difflib
    "golgonda": "Golconda: Inner Peace",
    "kine dominance": "Dominate Kine",
    "pentex=99 subversion": "Pentex Subversion",
    "soul gems": "Soul Gem of Etrius",
    "carver's meat packing": "Carver's Meat Packing and Storage",
    "alia, god=92s messenger": "Alia, God's Messenger",
    "truth of a 1000 lies": "Truth of a Thousand Lies",
    "krcg newspaper": "KRCG News Radio",
    "ps: istanbul": "Praxis Seizure: Istanbul",
    "2th tradition": "Second Tradition: Domain",
    "info superhighway": "Information Highway",
}
TYPE_ORDER = [
    "Master",
    "Conviction",
    "Action",
    "Action/Combat",
    "Action/Reaction",
    "Political Action",
    "Action Modifier",
    "Action Modifier/Reaction",
    "Action Modifier/Combat",
    "Combat",
    "Combat/Action",
    "Combat/Action Modifier",
    "Combat/Reaction",
    "Reaction",
    "Reaction/Action Modifier",
    "Reaction/Combat",
    "Power",
    "Equipment",
    "Ally",
    "Retainer",
    "Event",
]
BANNED = [
    # VEKN card lists are not up to date.
    "Anthelios, The Red Star",
    "Dramatic Upheaval",
    "Edge Explosion",
    "Kindred Restructure",
    "Lilith's Blessing",
    "Madness of the Bard",
    "Memories of Mortality",
    "Protect Thine Own",
    "The Return to Innocence",
    "Seeds of Corruption",
    "Succubus Club",
    "Temptation of Greater Power",
    "Cunctator Motion",
    "High Stakes",
    "Playing for Keeps",
]
